# import matplotlib as mpl
# mpl.use("Agg")
# import matplotlib.pyplot as plt

import os
import math
import pickle
import mxnet as mx
import numpy as np
import pandas as pd
from utils import metrics
from sklearn.model_selection import train_test_split as tts
from sklearn.metrics import roc_auc_score

from sklearn.svm import SVC
from scipy.stats import pearsonr

class DNP:

    def __init__(self, hidden_depth=[2], activation_func="RELU", reg=0, max_features=25, max_iter=25, lr=0.1, dropout_rate=0.6, dropout_iter=60, metrics_func="auc"):

        #DNP class
        #hidden_depth: the arthitecture of the multi-layer perceptron
        #activation_func: the activation function for the neuron.
        #reg: the regularization
        #max_features: The maximum number of selected features
        #max_iter: The maximum iterations of backpropagation.
        #lr: Learning Rate of BP
        #dropout_rate: the dropout rate for the multiple dropout
        #dropout_iter: the number of dropout for each multiple dropout.
        #metrics_func: the evaludation metrics

        self.hidden_depth = hidden_depth
        self.activation_func = activation_func

        #The list of selected features
        self.active_d = []
        #The list of unselected features
        self.inactive_d = []

        self.norm_style = None
        self.activation_func = activation_func
        self.metrics_func = metrics_func
        self.reg = reg

        self.max_features = max_features 
        self.max_iter = max_iter 
        self.lr = lr 

        self.dropout_rate = dropout_rate
        self.dropout_iter = dropout_iter

        self.net = None

    def fit(self, X, y, tst_X=None, tst_y=None, vd_X=None, vd_y=None):

        #DNP: fit
        #X and y: the training features and labels
        #tst_X and tst_y: the test features and labels.
        #vd_X and vd_y: the validation features and labels.

        assert X.shape[0] == len(y)  # check if the number of samples in X is the same as y

        #Add the bias term
        X = np.hstack([np.ones((X.shape[0], 1)), X])  # x = np.c_[np.ones((X.shape[0], 1)), X]

        if vd_X is None or vd_y is None:    
            vd_X = X.copy()
            vd_y = y.copy()
        else:
            vd_X = np.hstack([np.ones((vd_X.shape[0], 1)), vd_X])
        if tst_X is None or tst_y is None:
            tst_X = X.copy()
            tst_y = y.copy()
        else:
            tst_X = np.hstack([np.ones((tst_X.shape[0], 1)), tst_X])

        #####Make Neural Network
        if self.net == None:
            self.stack_net()

        ####Bind the neural network
        self.net_exec = self.net.simple_bind(ctx=mx.cpu(1), data=X.shape, grad_req="write")
        self.tst_exec = self.net.simple_bind(ctx=mx.cpu(1), data=tst_X.shape, grad_req="write")
        self.vd_exec = self.net.simple_bind(ctx=mx.cpu(1), data=vd_X.shape, grad_req="write")

        self.sgd_n = int(X.shape[0] * 1)
        self.sgd_exec = self.net.simple_bind(ctx=mx.cpu(1), data=(self.sgd_n, X.shape[1]), grad_req="write")

        ####Assign the names and arrays to correspondent variable
        net_name = self.net_exec.arg_dict
        arg_names = self.net.list_arguments()
        data_name = "data"
        label_name = "softmax_label"
        param_names = list(set(arg_names) - set([data_name, label_name]))

        self.param_idx = [i for i in range(len(arg_names)) if arg_names[i] in param_names]
        self.data_idx = [i for i in range(len(arg_names)) if arg_names[i] == data_name][0]
        self.label_idx = [i for i in range(len(arg_names)) if arg_names[i] == label_name][0]
        self.input_idx = [i for i in range(len(arg_names)) if arg_names[i] == "fc1_weight"][0]

        self.param_arrays = [self.net_exec.arg_arrays[i] for i in self.param_idx]
        self.lr_arrays = [np.ones(self.param_arrays[i].shape) * self.lr for i in range(len(self.param_arrays))]
        self.gdc_arrays = [np.zeros(self.param_arrays[i].shape) for i in range(len(self.param_arrays))]

        self.grad_arrays = [self.net_exec.grad_arrays[i] for i in self.param_idx]
        self.data_arrays = self.net_exec.arg_dict[data_name]
        self.label_arrays = self.net_exec.arg_dict[label_name]
        
        self.input_param = self.net_exec.arg_arrays[self.input_idx]
        self.input_grad = self.net_exec.grad_arrays[self.input_idx]

        self.tst_data_arrays = self.tst_exec.arg_dict[data_name]
        self.tst_label_arrays = self.tst_exec.arg_dict[label_name]
        self.tst_param = [self.tst_exec.arg_arrays[i] for i in self.param_idx]
        self.input_grad_tst = self.tst_exec.grad_arrays[self.input_idx]

        self.vd_data_arrays = self.vd_exec.arg_dict[data_name]
        self.vd_label_arrays = self.vd_exec.arg_dict[label_name]
        self.vd_param = [self.vd_exec.arg_arrays[i] for i in self.param_idx]
        self.input_grad_vd = self.vd_exec.grad_arrays[self.input_idx]
        
        self.sgd_data_arrays = self.sgd_exec.arg_dict[data_name]
        self.sgd_label_arrays = self.sgd_exec.arg_dict[label_name]
        self.sgd_param = [self.sgd_exec.arg_arrays[i] for i in self.param_idx]
        self.input_grad_sgd = self.sgd_exec.grad_arrays[self.input_idx]

        #Initially, only the bias is active
        self.active_d = [0]
        self.inactive_d = list(range(1, X.shape[1]))   # Added bias item, bias column + 51 * 7070
        self.grad_rescaling = 1.0 / X.shape[0]
        self.init_lr = self.lr

        self.grad_past = []

        #####Initialization
        self.vd_data_arrays[:] = vd_X
        self.vd_label_arrays[:] = vd_y
        self.data_arrays[:] = X
        self.label_arrays[:] = y
        
        #Initialization of the neural network
        #All parameter is initialized with the Xavier Initializer.
        initializer = mx.init.Xavier(factor_type="out", magnitude=2.34)
        for i in range(len(self.param_arrays)):
            self.grad_past.append(np.zeros(self.param_arrays[i].shape))

            if self.param_idx[i] == self.input_idx:
                empty_nd = mx.nd.zeros((self.param_arrays[i].shape[0], 1), mx.cpu(1))
                initializer(mx.init.InitDesc(arg_names[self.param_idx[i]]), empty_nd)
                tmp_param0 = np.zeros(self.param_arrays[i].shape)
                tmp_param0[:, 0] = empty_nd.asnumpy().flatten()
                self.param_arrays[i][:] = tmp_param0
            else:
                initializer(mx.init.InitDesc(arg_names[self.param_idx[i]]), self.param_arrays[i])

        ## added by Jinhong. Store the model parameter.
        param_select = {}

        #Begin to selecting and optimizing the DNP model.
        while (self.max_features == -1) or (len(self.active_d) <= self.max_features):
            #Begin to Train
            it_loss = []

            #The Adaptative decreasing learning rate
            lr = self.init_lr
            lr = [self.lr_arrays[i] / np.sqrt(1 + self.gdc_arrays[i]) for i in range(len(self.lr_arrays))]

            for it in range(self.max_iter):

                if len(self.active_d) == 1 and it >= 20:
                    break

                #Forward pass of DNP
                self.net_exec.forward(is_train=True)
                
                #Backward pass of DNP
                self.net_exec.backward()

                #Backpropagation update of DNP
                for i in range(len(self.param_arrays)):
                    if self.param_idx[i] == self.input_idx:
                        param = self.param_arrays[i].asnumpy()
                        param_cp = param.copy()
                        grad = self.grad_arrays[i].asnumpy()
                        self.gdc_arrays[i] += (grad * self.grad_rescaling) ** 2
                        self.gdc_arrays[i][:, self.inactive_d] = 0
                        param += - lr[i] * (grad * self.grad_rescaling + self.reg * param)
                        param_cp[:, self.active_d] = param[:, self.active_d] 
                        self.param_arrays[i][:] = param_cp
                    else:
                        grad = self.grad_arrays[i].asnumpy()
                        self.gdc_arrays[i] += (grad * self.grad_rescaling) ** 2
                        param = self.param_arrays[i].asnumpy()
                        param -= lr[i] * (grad * self.grad_rescaling + self.reg * param)
                        self.param_arrays[i][:] = param

                ####Output the loss
                self.net_exec.forward(is_train=False)
                tr_pred = self.net_exec.outputs[0].asnumpy()[:,1].flatten()

                #Test
                self.tst_data_arrays[:] = tst_X
                self.tst_label_arrays[:] = tst_y

                #Validation
                self.vd_data_arrays[:] = vd_X
                self.vd_label_arrays[:] = vd_y

                for i in range(len(self.param_arrays)):
                    self.param_arrays[i].copyto(self.tst_param[i])
                    self.param_arrays[i].copyto(self.vd_param[i])
        
                self.tst_exec.forward(is_train=False)
                self.tst_exec.backward()
                tst_pred = self.tst_exec.outputs[0].asnumpy()[:,1].flatten()

                self.vd_exec.forward(is_train=False)
                self.vd_exec.backward()
                vd_pred = self.vd_exec.outputs[0].asnumpy()[:,1].flatten()

            iter_active_d = len(self.active_d)
            past_active_d = self.active_d[:]
            max_grad = self.add_feature(X, y)

            print("current feature active features selected " )
            print(past_active_d )

            # Added by Jinhong Shi
            # uncomment the following lines if you want to check the weight matrix
            # prefix = "fc1_weights_AT_%d_selected_features (0 is the implecity bias b0 for FC1)" %(len(self.active_d)-1)
            # print(prefix)
            # print(self.param_arrays[0][:, past_active_d])

            param_select[str(past_active_d[-1])] = self.param_arrays[0][:, past_active_d].asnumpy()
        
            out_str = "#Select Features:%d, Select Feature ID:%d, Tr %s:%f, Vd %s:%f, Tst %s:%f" % (
             iter_active_d-1, past_active_d[-1],
                 self.metrics_func, metrics(y, tr_pred, self.metrics_func),
                 self.metrics_func, metrics(vd_y, vd_pred, self.metrics_func),
                 self.metrics_func, metrics(tst_y, tst_pred, self.metrics_func))

            print(out_str)


        # Calculate the activation potential for the selected 19 features (biase included)
        W = self.param_arrays[0][:, past_active_d].asnumpy() # 30 x 20 dim-->20 hidden neurons * 30 features
        b = self.param_arrays[1].asnumpy()   # 30 scalar  --> 20 scalars for 20 hidden neurons
        Xselected = X[:, past_active_d]

        # calculate average absolute activation potential P_{N_inpu}*{N_hiddle1}
        pij = np.zeros((self.max_features, self.hidden_depth[0]))
        for mtr in range(0,X.shape[0]):   # M = 116  training samples
            for j in range(0, self.hidden_depth[0]):  # jth hidden neural in 1st Hidden Layer
                for i in range(0,self.max_features):      # ith features (X training data)
                     aa = Xselected[mtr,i] * W[j, i] + b[j]
                     pij[i,j] += np.fabs(aa)

        pij = pij / X.shape[0]

        # Calculate the activation contribution of i to all hidden neurons
        cij = np.zeros((self.max_features, self.hidden_depth[0]))
        ci = np.zeros((self.max_features, 1))
        # for mtr in range(0, 116):  # M = 116  training samples
        for mtr in range(0, X.shape[0]):  # M = 116  training samples
            for j in range(0, self.hidden_depth[0]):  # jth hidden neural in 1st Hidden Layer
                for i in range(0, self.max_features):  # ith features (X training data)
                    aa = Xselected[mtr, i] * W[j, i] + b[j]  # 1-->mtr
                    cij[i, j] = aa / sum(pij[:, j])
            # calculate the contribution of i to all hidden neurons when training sample is mtr
            # only add up positive contribution
            cij[cij < 0] = 0
            ci += cij.sum(axis=1).reshape(ci.shape)  # sum over columns, sum of each row, each feature
        ci = ci / X.shape[0]
        
        for i in range(0, self.max_features):  # ith features (X training data)
            # print("feature %d positive contribution %f" %(self.active_d[i], sum(np.maximum(0, cij[i,:]))))
            print("feature %d positive contribution %f" % (self.active_d[i], ci[i,0]))   # ci is a matrix with 1 column

        self.tr_pred = tr_pred
        self.tst_pred = tst_pred
        self.vd_pred = vd_pred


    def stack_net(self):

        if isinstance(self.hidden_depth, int):
            print("The Unit length must be a list")
            exit()

        net = mx.symbol.Variable("data")
        net = mx.symbol.Dropout(data=net, p=0.3, name="dropout%d" % 0)

        for i in range(len(self.hidden_depth) - 1):
            net = mx.symbol.FullyConnected(data=net, name="fc%d" % (i+1), num_hidden=self.hidden_depth[i], no_bias=False)
            net = mx.symbol.Activation(data=net, name='%s%d' % (self.activation_func, i+1),            act_type=self.activation_func.lower())
            net = mx.symbol.Dropout(data=net, p=self.dropout_rate, name="dropout%d" % (i+1))

        net = mx.symbol.FullyConnected(data=net, name="fc%d" % (len(self.hidden_depth)), num_hidden=self.hidden_depth[-1])
        net = mx.symbol.SoftmaxOutput(data=net, name="softmax")
        self.net = net

    def add_feature(self, X, y):

        if len(self.inactive_d) == 0:
            return -1

    #   #Calculate gradient list
        for i in range(len(self.param_arrays)):
            self.param_arrays[i].copyto(self.sgd_param[i])

        sum_grad_norm = np.zeros(self.input_grad_sgd.shape[1])
        min_grad_norm = np.ones(self.input_grad_sgd.shape[1]) * 100000

        for i in range(self.dropout_iter):

            sgd_id = np.arange(X.shape[0])
            np.random.shuffle(sgd_id)

            self.sgd_data_arrays[:] = X[sgd_id[0:self.sgd_n], :]
            self.sgd_label_arrays[:] = y[sgd_id[0:self.sgd_n]]
            #self.sgd_data_arrays[:] = X
            #self.sgd_label_arrays[:] = y
            
            self.sgd_exec.forward(is_train=True)
            self.sgd_exec.backward()
            
            sgd_grad = self.input_grad_sgd.asnumpy()
           
            #######Gradient Shrinkage##########
            sign_grad = np.sign(sgd_grad)
            abs_grad = np.abs(sgd_grad)
            shrink_grad = abs_grad - self.reg 
            shrink_grad[shrink_grad <= 0] = 0
            sgd_grad = sign_grad * shrink_grad
            #######Gradient Shrinkage##########
            
            norm_sgd_grad = np.linalg.norm(sgd_grad, axis=0, ord=self.norm_style).flatten()
            #norm_sgd_sgd = norm_sgd_grad / np.sum(norm_sgd_grad)
            sum_grad_norm += norm_sgd_grad
            min_grad_norm = np.min(np.vstack([min_grad_norm, norm_sgd_grad]), axis=0)
            
        sum_grad_norm = sum_grad_norm / self.dropout_iter   
        inactive_norm_grad = min_grad_norm[self.inactive_d]
    
        inactive_id = np.argmax(inactive_norm_grad)
        
        arg_sort = np.argsort(self.active_d)

        correct_select = np.sum(np.array(self.active_d) <= 100)

        #Forward Add Feature
        #########################################
        choice_id = inactive_norm_grad.argsort()[-1::][::-1]
        
        self.active_d += (np.array(self.inactive_d)[choice_id]).tolist()
        for inactive_id in np.sort(choice_id)[::-1]:
            self.inactive_d.pop(inactive_id)
        
        initializer = mx.initializer.Xavier(factor_type="out", magnitude=2.34)
        empty_nd = mx.nd.zeros((self.input_param.shape[0], 1), mx.cpu(1))
        initializer(mx.init.InitDesc("fc1_weight"), empty_nd)

        init_input_param = self.input_param.asnumpy()
        init_input_param[:, self.active_d[-1]] = empty_nd.asnumpy().flatten()
        self.input_param[:] = init_input_param

        return inactive_norm_grad[inactive_id] * self.grad_rescaling * self.init_lr
